/*------------------------------------------------------------------------------
* sdrcorr.c: sdr receiver correlator
*
*          Copyright (C) 2010 by T.TAKASU, All rights reserved.
*
* options : -DCLUT  use carrier lookup table
*
* version : $Revision:$ $Date:$
* history : 2010/10/31 1.0  new
*-----------------------------------------------------------------------------*/
#include "sdr.h"
#include "mex.h"

#if defined(CLUT)||!defined(SSE2)

#define CDIV            32              /* carrier lookup table divided cycle */
#define CMASK           0x1F            /* carrier lookup table mask */
#define CSCALE          (1.0/16.0)      /* carrier lookup table scale (LSB) */

static short cost[CDIV]={0};            /* carrier lookup table cos(t) */
static short sint[CDIV]={0};            /* carrier lookup table sin(t) */

static void initcarr(void)              /* initialize carrier lookup table */
{
    int i;
    for (i=0;i<CDIV;i++) {
        cost[i]=(short)floor((cos(DPI/CDIV*i)/CSCALE+0.5));
        sint[i]=(short)floor((sin(DPI/CDIV*i)/CSCALE+0.5));
    }
}
#else

#define CSCALE          1.0

#endif

/*------------------------------------------------------------------------------
* without sse2 instructions
*-----------------------------------------------------------------------------*/
#if !defined(SSE2)

/* mix local carrier ---------------------------------------------------------*/
static void mixcarr(const char *data, int type, double ti, int n, double freq,
                    double phi0, short *I, short *Q)
{
    const char *p=data;
    double phi,ps;
    int index;
    
    if (!cost[0]) initcarr(); /* initialize local carrier table */
    
    phi=phi0*CDIV/DPI+0.5;
    phi-=floor(phi/CDIV)*CDIV; /* initial phase (0<=phi<CDIV) */
    ps=freq*CDIV*ti; /* phase step */
    
    if (type) { /* complex */
        for (;p<p+n*2;p+=2,I++,Q++,phi+=ps) {
            index=((int)phi)&CMASK;
            *I=cost[index]*p[0];
            *Q=cost[index]*p[1];
        }
    }
    else { /* real */
        for (;p<data+n;p++,I++,Q++,phi+=ps) {
            index=((int)phi)&CMASK;
            *I=cost[index]*p[0];
            *Q=sint[index]*p[0];
        }
    }
}
/* dot products (2x1) --------------------------------------------------------*/
static void dot_21(const short *a1, const short *a2, const short *b, int n,
                   double *dot1, double *dot2)
{
    const short *p1=a1,*p2=a2,*q=b;
    
    dot1[0]=dot2[0]=0.0;
    
    for (;p1<a1+n;p1++,p2++,q++) {
        dot1[0]+=(*p1)*(*q);
        dot2[0]+=(*p2)*(*q);
    }
}
/* dot products (2x2) --------------------------------------------------------*/
static void dot_22(const short *a1, const short *a2, const short *b1,
                   const short *b2, int n, double *dot1, double *dot2)
{
    const short *p1=a1,*p2=a2,*q1=b1,*q2=b2;
    
    dot1[0]=dot1[1]=dot2[0]=dot2[1]=0.0;
    
    for (;p1<a1+n;p1++,p2++,q1++,q2++) {
        dot1[0]+=(*p1)*(*q1);
        dot1[1]+=(*p1)*(*q2);
        dot2[0]+=(*p2)*(*q1);
        dot2[1]+=(*p2)*(*q2);
    }
}
/* dot products (2x3) --------------------------------------------------------*/
static void dot_23(const short *a1, const short *a2, const short *b1,
                   const short *b2, const short *b3, int n, double *dot1,
                   double *dot2)
{
    const short *p1=a1,*p2=a2,*q1=b1,*q2=b2,*q3=b3;
    
    dot1[0]=dot1[1]=dot1[2]=dot2[0]=dot2[1]=dot2[2]=0.0;
    
    for (;p1<a1+n;p1++,p2++,q1++,q2++,q3++) {
        dot1[0]+=(*p1)*(*q1);
        dot1[1]+=(*p1)*(*q2);
        dot1[2]+=(*p1)*(*q3);
        dot2[0]+=(*p2)*(*q1);
        dot2[1]+=(*p2)*(*q2);
        dot2[2]+=(*p2)*(*q3);
    }
}
/*------------------------------------------------------------------------------
* with sse2 instructions
*-----------------------------------------------------------------------------*/
#else /* SSE2 */

#if defined(CLUT)

/* carrier lookup table index ------------------------------------------------*/
#define CINDEX(x1,x2,dx,mask,index) { \
    __m128i _i1,_i2,_ind; \
    _i1=_mm_cvttpd_epi32(x1); \
    _i2=_mm_cvttpd_epi32(x2); \
    _ind=_mm_unpacklo_epi32(_i1,_i2); \
    _ind=_mm_and_si128(_ind,mask); \
    _mm_storeu_si128((void *)(index),_ind); \
    x1=_mm_add_pd(x1,dx); \
    x2=_mm_add_pd(x2,dx); \
}
/* mix local carrier ---------------------------------------------------------*/
extern void mixcarr(const char *data, int type, double ti, int n, double freq,
                    double phi0, short *I, short *Q)
{
    const char *p=data;
    double phi,ps;
    int index[8]={0};
    __m128d x1,x2,dx;
    __m128i mask;
    
    initcarr(); /* initialize local carrier table */
    
    phi=phi0*CDIV/DPI+0.5;
    phi-=floor(phi/CDIV)*CDIV; /* initial phase (0<=phi<CDIV) */
    ps=freq*CDIV*ti; /* phase step */
    
    x1=_mm_set_pd(phi+ps  ,phi);
    x2=_mm_set_pd(phi+ps*3,phi+ps*2);
    dx=_mm_set1_pd(ps*4);
    mask=_mm_set1_epi32(CMASK);
    
    if (type) { /* complex */
        for (;p<p+n*2;p+=8,I+=4,Q+=4) {
            
            CINDEX(x1,x2,dx,mask,index);
            I[0]=cost[index[0]]*p[0];
            I[1]=cost[index[2]]*p[2];
            I[2]=cost[index[1]]*p[4];
            I[3]=cost[index[3]]*p[6];
            Q[0]=cost[index[0]]*p[1];
            Q[1]=cost[index[2]]*p[3];
            Q[2]=cost[index[1]]*p[5];
            Q[3]=cost[index[3]]*p[7];
        }
    }
    else { /* real */
        for (;p<data+n;p+=4,I+=4,Q+=4) {
            
            CINDEX(x1,x2,dx,mask,index);
            I[0]=cost[index[0]]*p[0];
            I[1]=cost[index[2]]*p[1];
            I[2]=cost[index[1]]*p[2];
            I[3]=cost[index[3]]*p[3];
            Q[0]=sint[index[0]]*p[0];
            Q[1]=sint[index[2]]*p[1];
            Q[2]=sint[index[1]]*p[2];
            Q[3]=sint[index[3]]*p[3];
        }
    }
}
#else /* CLUT */

/* transform double x 8 to int16 x 8 and mask --------------------------------*/
#define DBLTOINT16(dst,src0,src1,mask) { \
    __m128i _a; \
    dst=_mm_cvttpd_epi32(src0); \
    _a =_mm_cvttpd_epi32(src1); \
    _a =_mm_slli_si128(_a,8); \
    dst=_mm_or_si128(dst,_a); \
    dst=_mm_and_si128(dst,mask); \
}
/* dst (int16 x 8) <- src (int8 x 8) -----------------------------------------*/
#define LOADINT8(dst,src,zero) { \
    __m128i _s; \
    dst=_mm_loadu_si128((__m128i *)(src)); \
    _s =_mm_cmplt_epi8(dst,zero); \
    dst=_mm_unpacklo_epi8(dst,_s); \
}
/* dst1,dst2 (int16 x 8) <- src (int8 x 16 interleaved) ----------------------*/
#define LOADINT8I(dst1,dst2,src) { \
    __m128i _d; \
    _d  =_mm_loadu_si128((__m128i *)(src)); \
    dst1=_mm_slli_epi16(_d,8); \
    dst1=_mm_srai_epi16(dst1,8); \
    dst2=_mm_srai_epi16(_d,8); \
}
/* add inclement to double x 8 -----------------------------------------------*/
#define ADDDBL(x1,x2,x3,x4,dx) { \
    x1=_mm_add_pd(x1,dx); x2=_mm_add_pd(x2,dx); \
    x3=_mm_add_pd(x3,dx); x4=_mm_add_pd(x4,dx); \
}
/* lookup table --------------------------------------------------------------*/
#define CLUT(dist,src,ind,ind_add,ind_sub) { \
    __m128i _a=_mm_setzero_si128(),_b; \
    _b=_mm_cmpeq_epi16(ind,ind_add); \
    _b=_mm_and_si128(_b,src); \
    _a=_mm_add_epi16(_a,_b); \
    _b=_mm_cmpeq_epi16(ind,ind_sub); \
    _b=_mm_and_si128(_b,src); \
    _a =_mm_sub_epi16(_a,_b); \
    _mm_storeu_si128((void *)(dist),_a); \
}
/* mix local carrier ---------------------------------------------------------*/
extern void mixcarr(const char *data, int type, double ti, int n, double freq,
                    double phi0, short *I, short *Q)
{
    const char *p=data;
    double phi,ps;
    __m128d x0,x2,x4,x6,dx;
    __m128i ind1,ind2,dat1,dat2,ind_cosp,ind_cosm,ind_sinp,ind_sinm;
    __m128i zero=_mm_setzero_si128();
    __m128i mask=_mm_set1_epi32(0x3);
    
    phi=phi0*4.0/DPI+0.5;
    phi-=floor(phi/4.0)*4.0; /* initial phase (0<=phi<4) */
    ps=freq*4.0*ti; /* phase step */
    
    x0=_mm_set_pd(phi+ps  ,phi);
    x2=_mm_set_pd(phi+ps*3,phi+ps*2);
    x4=_mm_set_pd(phi+ps*5,phi+ps*4);
    x6=_mm_set_pd(phi+ps*7,phi+ps*6);
    dx=_mm_set1_pd(ps*8);
    
    /* index of cost={1,0,-1,0},sint={0,1,0,-1} */
    ind_cosp=_mm_set1_epi16(0);
    ind_cosm=_mm_set1_epi16(2);
    ind_sinp=_mm_set1_epi16(1);
    ind_sinm=_mm_set1_epi16(3);
    
    if (type) { /* complex */
        for (;p<data+n*2;p+=16,I+=8,Q+=8) {
            
            /* ind1={(short)x7,(short)x6,...,(short)x0}&mask */
            DBLTOINT16(ind1,x0,x2,mask);
            DBLTOINT16(ind2,x4,x6,mask);
            ind1=_mm_packs_epi32(ind1,ind2);
            
            /* I=data_I.*cos(t), Q=data_Q.*cos(t) */
            LOADINT8I(dat1,dat2,p);
            CLUT(I,dat1,ind1,ind_cosp,ind_cosm);
            CLUT(Q,dat2,ind1,ind_cosp,ind_cosm);
            
            ADDDBL(x0,x2,x4,x6,dx);
        }
    }
    else { /* real */
        for (;p<data+n;p+=8,I+=8,Q+=8) {
            
            DBLTOINT16(ind1,x0,x2,mask);
            DBLTOINT16(ind2,x4,x6,mask);
            ind1=_mm_packs_epi32(ind1,ind2);
            
            /* I=data.*cos(t), Q=data.*sin(t) */
            LOADINT8(dat1,p,zero);
            CLUT(I,dat1,ind1,ind_cosp,ind_cosm);
            CLUT(Q,dat1,ind1,ind_sinp,ind_sinm);
            
            ADDDBL(x0,x2,x4,x6,dx);
        }
    }
}
#endif /* CLUT */

/* multiply and add ----------------------------------------------------------*/
#define MULADD(p,q,s) { \
    __m128i _x=_mm_loadu_si128((__m128i *)(p)); \
    __m128i _y=_mm_loadu_si128((__m128i *)(q)); \
    _y=_mm_madd_epi16(_y,_x); \
    s=_mm_add_epi32(s,_y); \
}
/* sum all -------------------------------------------------------------------*/
#define SUMALL(s,r) { \
    int _sum[4]; \
    _mm_storeu_si128((void *)_sum,s); \
    r=_sum[0]+_sum[1]+_sum[2]+_sum[3]; \
}
/* dot products (2x1) --------------------------------------------------------*/
static void dot_21(const short *a1, const short *a2, const short *b, int n,
                   double *dot1, double *dot2)
{
    const short *p1=a1,*p2=a2,*q=b;
    __m128i s1=_mm_setzero_si128();
    __m128i s2=_mm_setzero_si128();
    
    for (;p1<a1+n;p1+=8,p2+=8,q+=8) {
        MULADD(p1,q,s1);
        MULADD(p2,q,s2);
    }
    SUMALL(s1,dot1[0]);
    SUMALL(s2,dot2[0]);
}
/* dot products (2x2) --------------------------------------------------------*/
static void dot_22(const short *a1, const short *a2, const short *b1,
                   const short *b2, int n, double *dot1, double *dot2)
{
    const short *p1=a1,*p2=a2,*q1=b1,*q2=b2;
    __m128i s1=_mm_setzero_si128();
    __m128i s2=_mm_setzero_si128();
    __m128i s3=_mm_setzero_si128();
    __m128i s4=_mm_setzero_si128();
    
    for (;p1<a1+n;p1+=8,p2+=8,q1+=8,q2+=8) {
        MULADD(p1,q1,s1);
        MULADD(p1,q2,s2);
        MULADD(p2,q1,s3);
        MULADD(p2,q2,s4);
    }
    SUMALL(s1,dot1[0]);
    SUMALL(s2,dot1[1]);
    SUMALL(s3,dot2[0]);
    SUMALL(s4,dot2[1]);
}
/* dot products (2x3) --------------------------------------------------------*/
static void dot_23(const short *a1, const short *a2, const short *b1,
                   const short *b2, const short *b3, int n, double *dot1,
                   double *dot2)
{
    const short *p1=a1,*p2=a2,*q1=b1,*q2=b2,*q3=b3;
    __m128i s1=_mm_setzero_si128();
    __m128i s2=_mm_setzero_si128();
    __m128i s3=_mm_setzero_si128();
    __m128i s4=_mm_setzero_si128();
    __m128i s5=_mm_setzero_si128();
    __m128i s6=_mm_setzero_si128();
    
    for (;p1<a1+n;p1+=8,p2+=8,q1+=8,q2+=8,q3+=8) {
        MULADD(p1,q1,s1);
        MULADD(p1,q2,s2);
        MULADD(p1,q3,s3);
        MULADD(p2,q1,s4);
        MULADD(p2,q2,s5);
        MULADD(p2,q3,s6);
    }
    SUMALL(s1,dot1[0]);
    SUMALL(s2,dot1[1]);
    SUMALL(s3,dot1[2]);
    SUMALL(s4,dot2[0]);
    SUMALL(s5,dot2[1]);
    SUMALL(s6,dot2[2]);
}
#endif /* SSE2 */

/* correlation power function --------------------------------------------------
* correlation power function
* args   : char   *data     I   sampling data vector (n x 1 or 2n x 1)
*          int    dtype     I   sampling data type (0:real,1:complex)
*          double ti        I   sampling interval (s)
*          int    n         I   number of samples
*          double freq      I   carrier frequency (Hz)
*          int    sat       I   satellite number
*          int    ctype     I   code type
*          double crate     I   code chip rate (chip/s)
*          double *P        O   correlation power vector (n x 1)
* return : none
* notes  : signal=f(t)+i*g(t)
*            data={f(0),f(ti),f(2*ti),f(3*ti),f(3*ti),...   } : real
*            data={f(0),g(0),f(ti),g(ti),f(2*ti),g(2*ti),...} : complex
*          P=abs(ifft(fft(code).*fft(data.*e^(2*pi*freq*t*i)))).^2
*-----------------------------------------------------------------------------*/
extern void corrfunc(const char *data, int dtype, double ti, int n,
                     double freq, int sat, int ctype, double crate, double *P)
{
    cpx_t *data_cpx,*code_cpx;
    short *data_I,*data_Q,*code;
    
    if (!(data_I=(short *)malloc(sizeof(short)*n+32))||
        !(data_Q=(short *)malloc(sizeof(short)*n+32))||
        !(code  =(short *)malloc(sizeof(short)*n+32))||
        !(data_cpx=cpxmalloc(n))||
        !(code_cpx=cpxmalloc(n))) return;
    
    /* mix local carrier */
    mixcarr(data,dtype,ti,n,freq,0.0,data_I,data_Q);
    
    /* generate code */
    gencode(sat,ctype,0.0,ti*crate,n,code);
    
    /* transfer to complex */
    cpxcpx(data_I,data_Q,CSCALE,n,data_cpx);
    cpxcpx(code  ,NULL  ,CSCALE,n,code_cpx);
    
    /* convolution */
    cpxfft(data_cpx,n);
    cpxfft(code_cpx,n);
    cpxmul(data_cpx,code_cpx,1,n);
    cpxifft(data_cpx,n);
    
    /* correlation power function */
    cpxabs(data_cpx,n,P);
    
    free(data_I); free(data_Q); free(code);
    cpxfree(data_cpx);
    cpxfree(code_cpx);
}
/* correlator ------------------------------------------------------------------
* multiply sampling data and carrier (I/Q), multiply code (E/P/L), and integrate
* args   : char   *data     I   sampling data vector (n x 1 or 2n x 1)
*          int    dtype     I   sampling data type (0:real,1:complex)
*          double ti        I   sampling interval (s)
*          int    n         I   number of samples
*          double freq      I   carrier frequency (Hz)
*          double phi0      I   carrier initial phase (rad)
*          int    sat       I   satellite number
*          int    ctype     I   code type
*          double crate     I   code chip rate (chip/s)
*          double coff      I   code chip offset (chip)
*          int    s         I   correlator spacing (sample)
*          int    m         I   number of side-correlators
*          short  *I,*Q     O   correlation power I,Q
*                                 I={I_P,I_E1,I_L1,I_E2,I_L2,...,I_Em,I_Lm}
*                                 Q={Q_P,Q_E1,Q_L1,Q_E2,Q_L2,...,Q_Em,Q_Lm}
* return : none
* notes  : see above for data
*-----------------------------------------------------------------------------*/
extern void correlator(const char *data, int dtype, double ti, int n,
                       double freq, double phi0, int sat, int ctype,
                       double crate, double coff, int s, int m, double *I,
                       double *Q)
{
    short *data_I,*data_Q,*code_e,*code;
    int i;
    
    if (!(data_I=(short *)malloc(sizeof(short)*n+32))||
        !(data_Q=(short *)malloc(sizeof(short)*n+32))||
        !(code_e=(short *)malloc(sizeof(short)*(n+2*s*m+1)+32))) return;
    
    code=code_e+s*m;
    
    /* mix local carrier */
    mixcarr(data,dtype,ti,n,freq,phi0,data_I,data_Q);
    
    /* generate code */
    gencode(sat,ctype,coff+s*m*ti*crate,ti*crate,n+2*s*m+1,code_e);
    
    /* multiply code and integrate */
    if (m<=0) {
        dot_21(data_I,data_Q,code,n,I,Q);
    }
    else {
        dot_23(data_I,data_Q,code,code-s,code+s,n,I,Q);
    }
    for (i=1;i<m;i++) {
        dot_22(data_I,data_Q,code-(i+1)*s,code+(i+1)*s,n,I+1+i*2,Q+1+i*2);
    }
    for (i=0;i<1+2*m;i++) {
        I[i]*=CSCALE;
        Q[i]*=CSCALE;
    }
    free(data_I); free(data_Q); free(code_e);
}
