/*------------------------------------------------------------------------------
* sdrfft.c: sdr receiver fft functions
*
*          Copyright (C) 2010 by T.TAKASU, All rights reserved.
*
* options : -DFFTW   use fftw
*           -DMKLFFT use mkl dft with fftw interface
*           -DCUFFT  use cufft with cuda
*
* version : $Revision:$ $Date:$
* history : 2010/10/31 1.0  new
*-----------------------------------------------------------------------------*/
#include "sdr.h"

/*------------------------------------------------------------------------------
* FFTW or Intel MKL FFTW Interface
*-----------------------------------------------------------------------------*/
#if defined (FFTW) || (MKLFFT)

/* complex malloc ------------------------------------------------------------*/
extern cpx_t *cpxmalloc(int n)
{
    return (cpx_t *)fftwf_malloc(sizeof(cpx_t)*n);
}
/* complex free --------------------------------------------------------------*/
extern void cpxfree(cpx_t *cpx)
{
    fftwf_free(cpx);
}
/* generate complex vector ---------------------------------------------------*/
extern void cpxcpx(const short *real, const short *imag, double scale, int n,
                   cpx_t *cpx)
{
    int i;
    for (i=0;i<n;i++) {
        cpx[i][0]=real[i]*(float)scale;
        cpx[i][1]=imag?imag[i]*(float)scale:0.0f;
    }
}
/* complex fft ---------------------------------------------------------------*/
extern void cpxfft(cpx_t *cpx, int n)
{
    fftwf_plan p=fftwf_plan_dft_1d(n,cpx,cpx,FFTW_FORWARD,FFTW_ESTIMATE);
    fftwf_execute(p);
    fftwf_destroy_plan(p);
}
/* complex ifft --------------------------------------------------------------*/
extern void cpxifft(cpx_t *cpx, int n)
{
    fftwf_plan p=fftwf_plan_dft_1d(n,cpx,cpx,FFTW_BACKWARD,FFTW_ESTIMATE);
    fftwf_execute(p);
    fftwf_destroy_plan(p);
}
/* complex multiply ----------------------------------------------------------*/
extern void cpxmul(cpx_t *cpxa, const cpx_t *cpxb, int flag, int n)
{
    cpx_t cpxa_i;
    int i;
    if (flag) { /* a.*conj(b) */
        for (i=0;i<n;i++) {
            cpxa_i[0]=cpxa[i][0];
            cpxa_i[1]=cpxa[i][1];
            cpxa[i][0]=-cpxa_i[0]*cpxb[i][0]-cpxa_i[1]*cpxb[i][1];
            cpxa[i][1]= cpxa_i[0]*cpxb[i][1]-cpxa_i[1]*cpxb[i][0];
        }
    }
    else { /* a.*b */
        for (i=0;i<n;i++) {
            cpxa_i[0]=cpxa[i][0];
            cpxa_i[1]=cpxa[i][1];
            cpxa[i][0]=cpxa_i[0]*cpxb[i][0]-cpxa_i[1]*cpxb[i][1];
            cpxa[i][1]=cpxa_i[0]*cpxb[i][1]+cpxa_i[1]*cpxb[i][0];
        }
    }
}
/* complex square of absolute value ------------------------------------------*/
extern void cpxabs(const cpx_t *cpx, int n, double *s)
{
    int i;
    for (i=0;i<n;i++) {
        s[i]=cpx[i][0]*cpx[i][0]+cpx[i][1]*cpx[i][1];
    }
}
/*------------------------------------------------------------------------------
* CUFFT
*-----------------------------------------------------------------------------*/
#elif defined (CUFFT)

/* complex malloc ------------------------------------------------------------*/
extern cpx_t *cpxmalloc(int n)
{
    cpx_t *p;
    if (cudaMalloc((void **)p,sizeof(cpx_t)*n)!=CUFFT_SUCCESS) return NULL;
    return p;
}
/* complex free --------------------------------------------------------------*/
extern void cpxfree(cpx_t *cpx)
{
    cudaFree(cpx);
}
/* generate complex vector ---------------------------------------------------*/
extern void cpxcpx(const short *real, const short *imag, double scale, int n,
                   cpx_t *cpx)
{
    int i;
    for (i=0;i<n;i++) {
        cpx[i].x=real[i]*(float)scale;
        cpx[i].y=imag?imag[i]*(float)scale:0.0;
    }
}
/* complex fft ---------------------------------------------------------------*/
extern void cpxfft(cpx_t *cpx, int n)
{
    cufftHandpe p;
    cufftPlan1d(&p,n,CUFFT_C2C,1);
    cufftExecC2C(&p,cpx,cpx,CUFFT_FORWARD);
    cufftDestroy(p);
}
/* complex ifft --------------------------------------------------------------*/
extern void cpxifft(cpx_t *cpx, int n)
{
    cufftHandpe p;
    cufftPlan1d(&p,n,CUFFT_C2C,1);
    cufftExecC2C(&p,cpx,cpx,CUFFT_INVERSE);
    cufftDestroy(p);
}
/* complex multiply ----------------------------------------------------------*/
extern void cpxmul(const cpx_t *cpxa, const cpx_t *cpxb, int flag, int n)
{
    cpx_t cpxa_;
    int i;
    if (flag) { /* a.*conj(b) */
        for (i=0;i<n;i++) {
            cpxa_=cpxa;
            cpxa[i].x=-cpxa_.x*cpxb[i].x-cpxa_[i].y*cpxb[i].y;
            cpxa[i].y= cpxa_.x*cpxb[i].y-cpxa_[i].y*cpxb[i].x;
        }
    }
    else { /* a.*b */
        for (i=0;i<n;i++) {
            cpxa_=cpxa;
            cpxa[i].x=cpxa_.x*cpxb[i].x-cpxa_[i].y*cpxb[i].y;
            cpxa[i].y=cpxa_.x*cpxb[i].y+cpxa_[i].y*cpxb[i].x;
        }
    }
}
/* complex square of absolute value ------------------------------------------*/
extern void cpxabs(const cpx_t *cpx, int n, double *s)
{
    int i;
    for (i=0;i<n;i++) {
        s[i]=cpx[i].x*cpx[i].x+cpx[i].y*cpx[i].y;
    }
}
#endif

